function [Feature, PCsMap, SetFeataux]= SetFeat(X,sIND,y,paramstruct)
% SETFEAT: Feature extraction for set classification (to be used within
% SetFeatClass.m)
%
% [F PCsMap] = SetFeat(X,v,y) with inputs X, p x M matrix, consisting
%   of M columns of p-dim'l observations vectors and v, a row vector of
%   length M indicating the indices of sets by natural numbers 1, ..., N,
%   and y, a vector of length M indicating the class memberships.
%   The code SetFeat handles the binary classification AND multicategoty
%   classification.
%   Output F is a cell array of extracted features.
%          PCsMap includes auxillary variables for mapping of new
%          observations.
%
% [F PCsMap SetFeataux] = SetFeat(X,v,y) with output SetFeataux containing
%   all auxiliary information.
%
% [F PCsMap SetFeataux] = SetFeat(X,v,y,paramstruct), see below for paramstruct.
%
% Input: paramstruct - a Matlab structure of input parameters
%                    Use: "help struct" and "help datatypes" to
%                         learn about these.
%                    Create one, using commands of the form:
%
%        paramstruct = struct('field1',values1,...
%                             'field2',values2,...
%                             'field3',values3) ;
%
%                          where any of the following can be used,
%                          these are optional, misspecified values
%                          revert to defaults
%
%    fields           values
%
%    iperm            0  no permutation test. Tuning parameter r >= 1.
%                     1  (default) perform permutation test.
%
%    nperm            numeric (default 100) the number of permutations for
%                                           the permutation test.
%
%    sizealpha        numeric (default 0.05) the size of the permutation
%                                            test.
%
%    igraphic         0  (default) Show no graphical result
%                     1  Graphically show the result of tuning process in
%                     current figure (clears current figure).
%
%    r                numeric. Inclusion of this parameter skips the tuning
%                              process and use the given r. Information in
%                              the input argument y will not be used.
%
%    featureoption    0 (default) use mean and PC space
%                     1 use total PC variance (together with mean and PC
%                     space)
%
% by Sungkyu Jung, Feb 2013, June 2013. July 2013.
%
% See Also SetFeatMap, setFeatClass, setFeatPredict




if nargin < 3
    help SetFeat.m; return;
end


%% 1. preparation
[p Nn] = size(X);
% p : dimension where each observation lies
% Nn = sum_{i=1}^N n_i (total number of observations);
uniquesetindex=unique(sIND);
N = length(uniquesetindex);
nvec = zeros(N,1); % nvec: vector of set size (n_i)
for i = 1:N
    nvec(i)=sum(sIND == uniquesetindex(i));
end


R = min([p; nvec]);
% there are at most R = min(p, n_i) nontrivial components in all sets.


% set parameters for default setting
iperm = 1;         % use permutation
nperm = 100;       % number of permutation
sizealpha = 0.05;  % size of the test
igraphic = 0;      % show no result;
irset = false;          % no preset tuning parameter r
featureoption = 0;

if nargin > 3
    if isfield(paramstruct,'iperm') ;    %  then change to input value
        iperm = paramstruct.iperm;
    end ;
    
    if isfield(paramstruct,'nperm') ;    %  then change to input value
        nperm = paramstruct.nperm;
    end ;
    
    if isfield(paramstruct,'sizealpha') ;    %  then change to input value
        sizealpha = paramstruct.sizealpha;
    end ;
    
    if isfield(paramstruct,'igraphic');    %  then change to input value
        igraphic = paramstruct.igraphic;
        if iperm ~= 1 || irset
            igraphic = 0;
        end
    end ;
    
    if isfield(paramstruct,'r');
        irset = true;
        rset = paramstruct.r;
    end
    if isfield(paramstruct,'featureoption') ;    %  then change to input value
        featureoption = paramstruct.featureoption;
    end ;
end


if ~irset  % if no pre-selected r, then we use information on y
    if N == length(y);
        yulabel = unique(y);
        K = length(yulabel); % K = 2 if binary classification.
        iclass = zeros(N,K);
        Nsub = zeros(K,1);
        for k = 1:K
            tmpid = (y == yulabel(k));
            iclass(:,k) = tmpid;
            Nsub(k) = sum(tmpid);
        end
        iclass = logical(iclass);
        
        if K < 2
            disp('Warning from SetFeat.m: labels y contain one unique item');
            disp('Skip tuning process and set r = ceil(max(p,n)/2)');
            irset = true;
            rset = ceil(R/2);
        end
    else
        disp('Warning from SetFeat.m: length of labels y inconsistent with set indices');
        disp('Skip tuning process and set r = ceil(max(p,n)/2)');
        irset = true;
        rset = ceil(R/2);
    end
end

%% 2. Now compute the mean and covariance matrices
meanv = zeros(p,N);
dirv = zeros(p,R,N);
latentv = zeros(R,N);
for j = 1:N
    Xj = X(:,sIND == uniquesetindex(j));
    meanv(:,j) = mean(Xj,2);
    [PCdirections Singularvalues]=svd(Xj - repmat(meanv(:,j),1,nvec(j)),0);
    Singularvalues = diag(Singularvalues)/nvec(j);
    latentv(:,j) = Singularvalues(1:R);
    dirv(:,:,j) = PCdirections(:,1:R);
end

%% 3. Ecudliean Data Matrix (Zstar) by cMDS
% For each choice of r = 1, ... , R
% compute the Euclidean Data Matrix by Classical Multidimensional Scaling.

Zstar = cell(R,1);
D2array = cell(R,1);
xlsarray = cell(R,1);
xusarray = cell(R,1);
for r = 1:R
    % run if irset = false; ifirset = true, then run only if r= rset;
    if (irset && r == rset ) || ~irset
        
        % MDS i) obtain pairwise distances between PC1-r spaces;
        D2 = zeros(N); % squared pair-wise distance
        for i = 1:N
            for j = i+1:N
                D2(i,j) =  (SineMetric(dirv(:,1:r,i),dirv(:,1:r,j)))^2;
            end
        end
        pdist2Mat = D2 + D2'; % symmetric pairwise distance^2 matrix
        
        % MDS ii) obtain the inner product matrix
        C = eye(N) - 1/N*ones(N);
        B = -1/2* C * pdist2Mat * C;
        
        % MDS iii) Now find Euclidean Data Matrix
        [a b ]=eig(B);
        [xls inddescevalues]=sort(diag(b),'descend');
        intrinsicdimindicator = (xls > 1e-9);
        if sum(intrinsicdimindicator) == 0
            intrinsicdimindicator = [1 ; zeros(N-1,1)] == 1;
        end
        xls = diag(xls(intrinsicdimindicator));
        xus = a(:,inddescevalues(intrinsicdimindicator));
        
        % Save results for later use (mapping).
        Zstar{r} = sqrt(xls) * xus';
        D2array{r} = pdist2Mat;
        xlsarray{r} = xls;
        xusarray{r} = xus;
    end
    
end

%% Tuning by modified Hotelling's T and by permutation test.

if ~irset % there's no pre-determined r0; Tuning required.
    
    T = zeros(R,1);
    
    if K == 2
        % binary classification!!
        for r = 1:R
            Zstar1 = Zstar{r}(:,iclass(:,1)); %use group1id --> iclass(:,k), N1 -> Nsub(k)
            Zstar2 = Zstar{r}(:,iclass(:,2));
            difference = mean(Zstar1,2) - mean(Zstar2,2);
            diagSpooled = ((Nsub(1)-1)*var(Zstar1,0,2) + (Nsub(2)-1)*var(Zstar2,0,2))/N;
            T(r) = sum(difference.^2 ./ diagSpooled);
        end
        [maxT r0]=max(T);
        
        if iperm
            Trand = zeros(nperm,1);
            Trandall = zeros(nperm,R);
            for jperm = 1:nperm;
                rind = randperm(N);
                randomgroupid = zeros(N,K);
                for k = 1:K
                    randomgroupid(rind(iclass(:,k)),k) = 1;
                end
                randomgroupid = logical(randomgroupid);
                Trand_candidate = zeros(1,R);
                for r = 1:R
                    Zstar1 = Zstar{r}(:,randomgroupid(:,1));
                    Zstar2 = Zstar{r}(:,randomgroupid(:,2));
                    difference = mean(Zstar1,2) - mean(Zstar2,2);
                    diagSpooled = ((Nsub(1)-1)*var(Zstar1,0,2) + (Nsub(2)-1)*var(Zstar2,0,2))/N;
                    Trand_candidate(r) = sum(difference.^2 ./ diagSpooled);
                end
                Trandall(jperm,:) = Trand_candidate;
                Trand(jperm) = max(Trand_candidate);
            end
            emp_pvalue = mean(T(r0) <= Trand);
            ifeat = emp_pvalue <= sizealpha;
        else
            ifeat = true;
        end
        
    else % (if K == 2) then this is multicalss classification!
        for r = 1:R
            Trtmp = 0;
            for k = 1:K
                iclassk = iclass(:,k);
                iclassbutk = logical(sum(iclass(:,1:K ~=k),2));
                Nk = Nsub(k);
                Nbutk = sum(Nsub(1:K ~= k ));
                Zstar1 = Zstar{r}(:,iclassk); %use group1id --> iclass(:,k), N1 -> Nsub(k)
                Zstar2 = Zstar{r}(:,iclassbutk);
                difference = mean(Zstar1,2) - mean(Zstar2,2);
                diagSpooled = ((Nk-1)*var(Zstar1,0,2) + (Nbutk-1)*var(Zstar2,0,2))/N;
                Trtmp = Trtmp + sum(difference.^2 ./ diagSpooled);
            end
            T(r) = Trtmp;
        end
        [maxT r0]=max(T);
        
        if iperm
            Trand = zeros(nperm,1);
            Trandall = zeros(nperm,R);
            for jperm = 1:nperm;
                
                rind = randperm(N);
                randomiclass = zeros(N,K);
                for k = 1:K
                    randomiclass(rind(iclass(:,k)),k) = 1;
                end
                
                Trand_candidate = zeros(1,R);
                for r = 1:R
                    Trtmp = 0;
                    for k = 1:K
                        randomiclassk = logical(randomiclass(:,k));
                        randomiclassbutk = logical(sum(randomiclass(:,1:K ~=k),2));
                        Nk = Nsub(k);
                        Nbutk = sum(Nsub(1:K ~= k ));
                        Zstar1 = Zstar{r}(:,randomiclassk); %use group1id --> iclass(:,k), N1 -> Nsub(k)
                        Zstar2 = Zstar{r}(:,randomiclassbutk);
                        difference = mean(Zstar1,2) - mean(Zstar2,2);
                        diagSpooled = ((Nk-1)*var(Zstar1,0,2) + (Nbutk-1)*var(Zstar2,0,2))/N;
                        Trtmp = Trtmp + sum(difference.^2 ./ diagSpooled);
                    end
                    Trand_candidate(r) = Trtmp;
                end
                Trandall(jperm,:) = Trand_candidate;
                Trand(jperm) = max(Trand_candidate);
            end
            emp_pvalue = mean(T(r0) <= Trand);
            ifeat = emp_pvalue <= sizealpha;
        else
            ifeat = true;
        end
        
        
        
    end
else % We then have predetermined tuning parameter r;
    r0 = rset;
    if r0 == 0;
        ifeat = false;
    else
        ifeat = true;
    end
end

%% Prepare output values
if featureoption == 1 && ifeat 
    % then use the totalPCvariance
    % if ifeat == 0 then use of totalPCvariance is depreciated.
    Feature.mean = [meanv;
    sum(latentv(1:r0,:),1)];
    PCsMap.featureoption = 1;
    disp('Message From SetFeat.m: total PCvariance is added to a feature');
else
    Feature.mean = meanv;
    PCsMap.featureoption = 0;
end

PCsMap.ifeat = ifeat;

if ifeat
    Feature.PCspace = Zstar{r0};
    Feature.PCvariance = latentv(1:r0,:);
    PCsMap.dirv = dirv(:,1:r0,:);
    PCsMap.D2 = D2array{r0};
    PCsMap.xls = xlsarray{r0};
    PCsMap.xus = xusarray{r0};
    PCsMap.r0 = r0;
    PCsMap.scale = mean(sum(Feature.PCvariance,1));
    
else % then do not use any covariance feature
    
    Feature.PCspace = [];
    Feature.PCvariance = [];
    PCsMap.scale = 0;
    
end


if nargout == 3  % then return all auxiliary information
    if irset
        SetFeataux.r0 = r0;
    else
        SetFeataux.r0 = r0;
        SetFeataux.R = R;
        SetFeataux.dirv = dirv;
        SetFeataux.D2 = D2array;
        SetFeataux.xls = xlsarray;
        SetFeataux.xus = xusarray;
        SetFeataux.Zstar = Zstar;
        SetFeataux.T = T;
        SetFeataux.K = K;
        if iperm
            SetFeataux.pvalue = emp_pvalue;
        end
    end
end


if igraphic % then show graphical results on the current fgure;
    
    disp([' Message from SetFeat.m: Using the training data with ' num2str(K) ' classes, ']);
    disp(['  i) The PC space dimension r is tuned to ' num2str(r0) ]);
    disp([' ii) Permutation test results in empricial p-values of ' num2str(emp_pvalue) ' for H0: "no feature". ']);
    if emp_pvalue <= sizealpha
        disp(['      With size ' num2str(sizealpha) ', the test rejects H0 and tunned r = ' num2str(r0)]);
    else
        disp(['      With size ' num2str(sizealpha) ', the test accepts H0 and tunned r = 0']);
    end
    
    clf;
    subplot(2,2,1);
    histJ(Trand);
    if K == 2;
        xlabel('T');
    else
        xlabel(['T from ' num2str(K) '-categoty classification']);
    end
    ylabel('kernel density estimate of Tb');
    hold on;
    yaxislim = get(gca,'ylim');
    plot([T(r0),T(r0)],yaxislim/2,'k');
    title({['observed T = ' num2str(T(r0))],...
        ['with empirical p-value = '  num2str(emp_pvalue) ]})
    
    subplot(2,1,2);
    hold on;
    for iperm = 1:nperm
        plot(1:R,Trandall(iperm,:),'-k','Linewidth',0.5)
    end
    plot(1:R,T,'-r','Linewidth',0.5)
    ylabel('T(r)');
    xlabel('r');
    xlim([0.5, R+0.5])
    title(['observed T(r) vs ' num2str(nperm) ' permuted Tb'])
    
    
    subplot(2,2,2);
    distmax = max(max(pdist2Mat));
    a = pdist2Mat + 1e+10*eye(size(pdist2Mat));
    distmin = min(min(a));
    clims = sqrt([distmin distmax]);
    imagesc(sqrt(pdist2Mat),clims); colorbar;
    title('pairwise distances');
end














end


function dist = SineMetric(u1,u2)
% SINEMETRIC between two subspaces
%       with dim p and dim q (out of dim d - the whole space)
%       u1 = d x p matrix of orthogonal frames
%       u2 = d x q matrix of orthogonal frames
%       dist gives the metric

a = u2'*u1;
n = max(size(u1,2),size(u2,2));
dist = real(sqrt(n-trace(a*a')));
end


function [kdeh xgrid] = histJ(vec)
% HISTSJ smooth histogram by simple kde.
[r c]=size(vec);
if r > 1 && c >1 ; disp('histSJ.m only handles vector input.');quit;end
if r > 1 && c == 1; % then convert into row vector;
    vec = vec';
end
n = length(vec);

[kde xgrid]=kdeSIMPLE(vec);
heights = linspace(max(kde),2*max(kde),n);
plot(xgrid, kde); hold on;
scatter(vec,heights,5);
hold off; xlabel( inputname(1))
ylabel('kde')

end

%% simple kde with an input as row vector.
function [kdeh xgrid]=kdeSIMPLE(rowvec)
%rowvec = imat(1,:);
n = length(rowvec);
xgrid = linspace(min(rowvec),max(rowvec),201);
nbin = length(xgrid);
dsd = std(rowvec) ; diqr = iqr(rowvec) ;
a = min([dsd; (diqr / 1.34)]) ; %  Using Silverman's notation
if a == 0;
    a = max([dsd; (diqr / 1.34); eps]);
end
h = .9 * a * n^(-1/5) ; % Silverman's eqn. 3.31 'rule of thumb'

kdeh = repmat((rowvec ./ h),nbin,1) - repmat((xgrid' ./ h),1,n) ;
%  efficient way to divide all dif's by h
%  variable name "kde" is used to avoid creating too many biggies
kdeh = exp(-(kdeh .^2) / 2) ;
%  exponential part of Gaussian density
kdeh = sum(kdeh,2)' ;
%  sum part of kde, and make result a column vector
kdeh = kdeh / (n * h * sqrt(2 * pi)) ;
%  normalize, and mult by Gaussain density constant

end
