function [delta, PCsMap, SetFeataux]= SetFeatClass(X,sIND,y,paramstruct)
% SETFEATCLASS : Train set classification rule using the PCspace features
%
% [delta PCsMap] = SetFeatClass(X,v,y) with inputs X, p x M matrix, consisting
%   of M columns of p-dim'l observations vectors and v, a row vector of
%   length M indicating the indices of sets by natural numbers 1, ..., N,
%   and y, a vector of length M indicating the class memberships.
%   This code SetFeatClass handles the binary AND multiclassification only.
%   output :
%       delta is a cell array of function handles, to be used as
%          \hat{y} = argmax_k delta(X*){k}.
%       PCsMap includes auxillary variables for mapping of new
%          observations.
%
% [delta PCsMap SetFeataux] = SetFeatClass(X,v,y) with output SetFeataux containing
%   all auxiliary information.
%
% [delta PCsMap] = SetFeatClass(X,v,y,paramstruct), see below for paramstruct.
%
% Input: paramstruct - a Matlab structure of input parameters
%                    Use: "help struct" and "help datatypes" to
%                         learn about these.
%                    Create one, using commands of the form:
%
%        paramstruct = struct('field1',values1,...
%                             'field2',values2,...
%                             'field3',values3) ;
%
%                          where any of the following can be used,
%                          these are optional, misspecified values
%                          revert to defaults
%
%    fields           values
%
%    iperm            0  no permutation test. Tuning parameter r >= 1.
%                     1  (default) perform permutation test.
%
%    nperm            numeric (default 100) the number of permutations for
%                                           the permutation test.
%
%    sizealpha        numeric (default 0.05) the size of the permutation
%                                            test.
%
%    igraphic         0  (default) Show no graphical result
%                     1  Graphically show the result of tuning process in
%                     current figure (clears current figure).
%
%    r                numeric. Inclusion of this parameter skips the tuning
%                              process and use the given r. Information in
%                              the input argument y will not be used.
%
%
%    scale            numeric (default = mean eigenvalues) 
%                     to be used as a scaler to make
%                     PC space feature to be commensurate with mean
%                     features. [Needs more discussion on this]
%
%    iclassifier      option for classifier in the feature space
%                     'LDA' (default) Linear Discriminant Analysis
%                     'QDA' Quadratic Discriminant Analysis
%                     'SVM' Support Vector Machine 
%                     'DWD' Distance Weighted Discrimination 
%                     'MDEB' Minimum distance Empirical Bayes (Srivastava &
%                     Kubokawa 2007) linear discriminant analysis
%                     'YA' Yata & Aoshima's modified LDA
% 
%                     Note: option 'SVM' and 'DWD' needs SVM1XQ.m and DWD2XQ.m
%                     For multiclass, use only 'LDA' or 'QDA'.
%
%    classifieroption numeric (default 0.01) any additional parameter for
%                     classifier.
%    
%    featureoption    0 (default) use mean and PC space
%                     1 use total PC variance (together with mean and PC
%                     space) 
%                     (this option is directly carried over to SetFeat.m)
%
%    Feature          if this is present, use Feature (for simulation).
%                     User must provide 'PCsMap'.
%    PCsMap           if this is present, use PCsMap (for simulation).
%
% By Sungkyu Jung, Feb-Nov 2013.
% Fixed: scale parameter 'c' is now computed in SetFeat and carried over 
%        in the struct PCsMap 6/11/2013
% Fixed: multicalss LDA and QDA classification. 6/26/2013
% Fixed: 'featureoption' added. 7/1/2013
% Fixed: classifiers 'MDEB' and 'YA' added. 11/20/2013
%
% See also SetFeat, SetFeatMap, setFeatPredict

if nargin < 3
    help SetFeatClass.m; return;
end

if nargin > 3;
    if isfield(paramstruct,'Feature');
        F = paramstruct.Feature;
        PCsMap = paramstruct.PCsMap;
        SetFeataux =[]; 
    else
    [F, PCsMap, SetFeataux]= SetFeat(X,sIND,y,paramstruct);
    end
else    
    [F, PCsMap, SetFeataux]= SetFeat(X,sIND,y);
end
 
scale = PCsMap.scale; 
iclassifier = 'LDA';
classifieroption = 0.01; 

if nargin > 3
    if isfield(paramstruct,'scale') ;    %  then change to input value
        scale = paramstruct.scale;
    end ;
    
    if isfield(paramstruct,'iclassifier') ;    %  then change to input value
        iclassifier = paramstruct.iclassifier;
    end ;
    
    if isfield(paramstruct,'classifieroption') ;    %  then change to input value
        classifieroption = paramstruct.classifieroption;
    end ; 
end


% check if multiclass discrimination;
label = unique(y);
K = length(label);
if K > 2
    if (strcmp(iclassifier,'LDA') || strcmp(iclassifier,'QDA'))
    % mulclass AND (LDA or QDA) is okay to proceed;
    % no implementation for multiclass SVM and DWD.
    % good to go;
    else
        iclassifier = 'LDA';
        disp(['Warning from SetFeatClass.m: option iclassifier = ' iclassifier ' is not supported' ]);
        disp(['Warning from SetFeatClass.m: for multiclass classification. Set iclassifier = LDA.' ]);
    end
end
  
 
% Prediction is hat{y} = argmax_k delta(newF.mean,newF.PCspace){k}

switch iclassifier
    case 'LDA'
        X = [F.mean ;F.PCspace*scale]';
        [N d]=size(X);
        nvec = zeros(K,1); % K-vector of class sizes
        muvec = zeros(d,K); % each column is the class-wise mean
        Sw = zeros(d,d); %With-in (common) covariance matrix
        for k = 1:K;
            nvec(k) = sum(y==label(k));
            muvec(:,k) = mean(X(y==label(k),:));
            Sw = Sw + cov(X(y==label(k),:))*(nvec(k)-1)/N;
        end
        Sw = Sw+classifieroption*eye(d);
        
        offset = zeros(1,K);
        ldadir = zeros(d,K);
           for k = 1:K;
             ldadir(:,k) =  pinv(Sw)*muvec(:,k);
             offset(k) = muvec(:,k)'*ldadir(:,k);
           end  
        
        delta = cell(K,1);
        for k = 1:K
            delta{k} = @(x,y) [x ; y*scale]'*ldadir(:,k) - (.5 * offset(k)) +log(nvec(k)); 
        end 
        
    case 'QDA'
        X = [F.mean ;F.PCspace*scale]';
        [N d]=size(X);
        nvec = zeros(K,1); % K-vector of class sizes
        muvec = zeros(d,K); % each column is the class-wise mean
        Sigmaarry = zeros(d,d,K); % covariance matrix for each class
        invSigmaarry = zeros(d,d,K); % covariance matrix for each class
        sumlogevalue = zeros(K,1);
        for k = 1:K
            nvec(k) = sum(y==label(k));
            muvec(:,k) = mean(X(y==label(k),:));
            Sigmaarry(:,:,k) = cov(X(y==label(k),:))+classifieroption*eye(d);
            sumlogevalue(k)=sum(log(eig(Sigmaarry(:,:,k))));
            if isinf(sumlogevalue(k))
                disp('Warning: Nonsingular covariance matrix: increase the sample size or add L2penalty');
                disp('       : Adding 1e-5 for numerical stability')
                Sigmaarry(:,:,k) = Sigmaarry(:,:,k) + (1e-5) * eye(d);
                sumlogevalue(k)=sum(log(eig(Sigmaarry(:,:,k))));
            end
            invSigmaarry(:,:,k) = inv(Sigmaarry(:,:,k));
        end
        
        delta = cell(K,1);
        for k = 1:K
            delta{k} = @(x,y) -1/2*sumlogevalue(k) ...
                - 1/2*diag(([x ; y*scale]-repmat(muvec(:,k),1,size(x,2)))' ...
                * (invSigmaarry(:,:,k) * ...
                ([x ; y*scale]-repmat(muvec(:,k),1,size(x,2))) ))...
                + log(nvec(k)/N);
        end 
    
    case 'SVM'
        X = [F.mean ;F.PCspace*scale];
        [w, b] = SVM1XQ(X(:,y==label(1)),X(:,y==label(2)));
        delta = cell(2,1);
        delta{1} = @(x,y) (w'*[x ; y*scale] + b > 0)' ;
        delta{2} = @(x,y) (w'*[x ; y*scale] + b <= 0)' ; 
     
    case 'DWD'
        X = [F.mean ;F.PCspace*scale];
        [w, b] = DWD2XQ(X(:,y==label(1)),X(:,y==label(2)));
        delta = cell(2,1);
        delta{1} = @(x,y) (w'*[x ; y*scale] + b > 0)' ;
        delta{2} = @(x,y) (w'*[x ; y*scale] + b <= 0)' ;    
    
    case 'MDEB'
        X = [F.mean ;F.PCspace*scale]';
        [N d]=size(X);
        nvec = zeros(K,1); % K-vector of class sizes
        muvec = zeros(d,K); % each column is the class-wise mean
        Sw = zeros(d,d); %With-in (common) covariance matrix
        for k = 1:K;
            nvec(k) = sum(y==label(k));
            muvec(:,k) = mean(X(y==label(k),:));
            Sw = Sw + cov(X(y==label(k),:))*(nvec(k)-1)/N;
        end
        
        Sw = Sw+(trace(Sw)/min(N,d))*eye(d);
        
        offset = zeros(1,K);
        ldadir = zeros(d,K);
           for k = 1:K;
             ldadir(:,k) =  pinv(Sw)*muvec(:,k);
             offset(k) = muvec(:,k)'*ldadir(:,k);
           end  
        
        delta = cell(K,1);
        for k = 1:K
            delta{k} = @(x,y) [x ; y*scale]'*ldadir(:,k) - (.5 * offset(k)) +log(nvec(k)); 
        end  
    case 'YA' 
        X = [F.mean ;F.PCspace*scale]';
        [N d]=size(X);
        nvec = zeros(K,1); % K-vector of class sizes
        muvec = zeros(d,K); % each column is the class-wise mean
        Sw = zeros(d,d); %With-in (common) covariance matrix
        for k = 1:K;
            nvec(k) = sum(y==label(k));
            muvec(:,k) = mean(X(y==label(k),:));
            Sw = Sw + cov(X(y==label(k),:))*(nvec(k)-1)/N;
        end
        [V,D] = eig(Sw);
        lambda = sort(diag(D),'descend'); 
        omega = sum(lambda)/max([sqrt(d)*N^(1/4), min(N,d)]);
        lambdanew = omega*zeros(d,1);
        for j = 1:min(d,(N-1));
            lambdanew(j) = max(lambda(j), omega* (lambda(j)/(lambda(j)-sum(lambda(j+1:end))/(N-j))));
        end
        Sw = V*diag(lambdanew)*V'; 
        
        offset = zeros(1,K);
        ldadir = zeros(d,K);
           for k = 1:K;
             ldadir(:,k) =  pinv(Sw)*muvec(:,k);
             offset(k) = muvec(:,k)'*ldadir(:,k);
           end  
        
        delta = cell(K,1);
        for k = 1:K
            delta{k} = @(x,y) [x ; y*scale]'*ldadir(:,k) - (.5 * offset(k)) +log(nvec(k)); 
        end 
        
    otherwise
        disp('message from SetFeatClass.m');
        disp(' Other options have not yet implemented. Choose iclassifier = LDA');
end


PCsMap.label = label;


