function Feature = SetFeatMap(varargin)
% SETFEATMAP : map new set into the feature space found by Set Feature
% Selection (done by SetFeat.m). (To be called within SetFeatPredict.m)
%
% Feature = SetFeatMap(A,PCsMap) returns the Feature vector of the p x n 
% matrix A consisting of columns of observations in a set.
% PCsMap is a Matlab structure, and is the output of SetFeat.
%
% Feature = SetFeatMap(A,v,PCsMap) returns the Feature vector of A,
% where A is a collection of many sets indicated by vector 'v' of set indices.
%
% By Sungkyu Jung, Feb 2013. June 2013. July 2013.
%
% See also SetFeat, setFeatClass, setFeatPredict

if nargin < 2
    help SetFeatMap.m; return;
end

if nargin == 2;
    X = varargin{1};
    [p Nn]=size(X);
    sIND = ones(1,Nn);
    PCsMap = varargin{2};
else
    X = varargin{1};
    sIND = varargin{2};
    PCsMap = varargin{3};
    [p Nn]=size(X);
end
 
%% Now compute the mean feature;
 

if PCsMap.ifeat  % then extract covariance features
    
    uniquesetindex=unique(sIND);
    N = length(uniquesetindex);
    nvec = zeros(N,1); % nvec: vector of set size (n_i)
    for i = 1:N
        nvec(i)=sum(sIND == uniquesetindex(i));
    end
    
    r0 = PCsMap.r0; % the tuned dimension.
    meanv = zeros(p,N);
    dirv = zeros(p,r0,N);
    latentv = zeros(r0,N);
    for j = 1:N
        Xj = X(:,sIND == uniquesetindex(j));
        meanv(:,j) = mean(Xj,2);
        [PCdirections Singularvalues]=svd(Xj - repmat(meanv(:,j),1,nvec(j)));
        Singularvalues = diag(Singularvalues);
        if length(Singularvalues) < r0
            latentv(:,j)  = [Singularvalues ; zeros(r0 - length(Singularvalues),1)];
        else
            latentv(:,j) = Singularvalues(1:r0)/nvec(j);
        end
        dirv(:,:,j) = PCdirections(:,1:r0);
    end
    
    [ptmp ktmp Ntrain] = size(PCsMap.dirv);
    [tmp intrinsicdim]=size(PCsMap.xus);
    Xneweig = zeros(intrinsicdim,N);
    
    if intrinsicdim > 0
        D2dd = sum(sum(PCsMap.D2))/(Ntrain^2);
        D2dj = sum(PCsMap.D2)'/Ntrain;
        for i = 1:N;
            D2new = zeros(Ntrain,1);
            for j = 1:Ntrain
                D2new(j) = (SineMetric(PCsMap.dirv(:,:,j),dirv(:,:,i)))^2;
            end
            D2di = mean(D2new);
            b = -1/2*(D2new - D2dj - D2di + D2dd);
            Xneweig(:,i) = sqrt(pinv(PCsMap.xls)) * PCsMap.xus' * b;
        end
    end
    
    Feature.mean = meanv;
    Feature.PCspace = Xneweig;
    Feature.PCvariance = latentv;
    
    if PCsMap.featureoption == 1
        Feature.mean = [meanv ; sum(latentv)];
    end
    
else    % PCsMap.ifeat == 0, thus no need to extract cov. features.
    
    uniquesetindex=unique(sIND);
    N = length(uniquesetindex);
    % N : number of sets
    meanv = zeros(p,N);
    for i = 1:N
        Xi = X(:,sIND == uniquesetindex(i));
        meanv(:,i) = mean(Xi,2);
    end
    
    Feature.mean = meanv;
    Feature.PCspace = [];
    Feature.PCvariance = [];
    
end



end


function dist = SineMetric(u1,u2)
% SINEMETRIC between two subspaces
%       with dim p and dim q (out of dim d - the whole space)
%       u1 = d x p matrix of orthogonal frames
%       u2 = d x q matrix of orthogonal frames
%       dist gives the metric

a = u2'*u1;
n = max(size(u1,2),size(u2,2));
dist = real(sqrt(n-trace(a*a')));
end
