function [ldadir, cutpoint, pred, errorrate] = ldasetclass(datacol, sIND, y, paramstruct)
% LDASETCLASS : Fisher's Linear Discriminatation Analysis for Set
% Classification, with two options 'MV' (majority voting) and 'WV'
% (Weighted Voting by exchanging the order of sum and sign) for binary
% classification.
%
% [ldadir, cutpoint] = ldasetclass(C, sIND, y) returns the LDA rule trained
% by the dataset (X,sIND, y).  
%
% [ldadir, cutpoint, pred, errorrate] = ldasetclass(X, sIND, y,
% paramstruct) returns predicted label for testing data and errorrate, see 
% below for paramstruct. 
%
%  
% Input: paramstruct - a Matlab structure of input parameters
%                    Use: "help struct" and "help datatypes" to
%                         learn about these.
%                    Create one, using commands of the form:
%
%        paramstruct = struct('field1',values1,...
%                             'field2',values2,...
%                             'field3',values3) ;
%
%                          where any of the following can be used,
%                          these are optional, misspecified values
%                          revert to defaults
%
%    fields           values
%
%    classifier       'WV' (default) use the weighted voting
%                     'MV'   use the majority voting
%
%    classifieroption numeric (default 0.01) any additional parameter for
%                     classifier.
%
%    TestData         Test Data set (similar structure to X)
%
%    Testsetindex     Test Set index (similar structure to sIND)
%
%    TestLabel        Test group labels (similar structure to y)
%    
%
%
% By Sungkyu Jung, Feb 2013.
%
% See also SetFeat, SetFeatMap, setFeatClass, SetFeatPredict

if nargin < 3
    help ldasetclass.m; return;
end

% set default parameters
classifier = 'WV'; 
classifieroption = 0.01; % L2 penalty for LDA
ireturn_errorrate = 0;
ireturn_pred = 0;
pred = [];
errorrate = [];

if nargin > 3
    
    if isfield(paramstruct,'classifier') ;    %  then change to input value
        classifier = paramstruct.classifier;
    end ;
    
    if isfield(paramstruct,'classifieroption') ;    %  then change to input value
        classifieroption = paramstruct.classifieroption;
    end ;
    
    if isfield(paramstruct,'TestData') && isfield(paramstruct,'Testsetindex');    %  then change to input value
        datacolN = paramstruct.TestData;
        sINDN = paramstruct.Testsetindex;
        ireturn_pred = 1;
    end ;
    
    if isfield(paramstruct,'TestLabel');    %  then change to input value
        ynew = paramstruct.TestLabel;
        ireturn_errorrate = 1;
    end ;        
end
 
yunique = unique(y);

N = max(sIND);
Nn = size(datacol,2);
yind = zeros(1,Nn);
for i = 1:N
    yind(sIND == i) = y(i);
end

% Run lda for aggregated data;
[ldadir, cutpoint] = lda(datacol,yind,classifieroption);

if ireturn_pred == 1
    individualresult = ldadir'*datacolN - cutpoint ;
    Ntest = max(sINDN);
    
    pred = zeros(Ntest,1);
    
    switch classifier
        case 'MV'
            
            for i = 1:Ntest;
                if sum( sign(individualresult(sINDN == i))) > 0
                    pred(i) = yunique(1);
                else
                    pred(i) = yunique(2);
                end
            end
            
        case 'WV'
            
            for i = 1:Ntest;
                if sign( sum(individualresult(sINDN == i))) > 0
                    pred(i) = yunique(1);
                else
                    pred(i) = yunique(2);
                end
            end
    end
    if ireturn_errorrate == 1
    errorrate = mean(ynew ~= pred);
    end
end


 