function [delta, pred, errorrate, confusionmat] = setclassmulti(datacol, sIND, y, paramstruct)
% SETCLASSMULTI : Simple Discriminatation Analysis for Set
% Classification, with two options ('MV' (majority voting) and 'WV'
% (Weighted Voting)) and four choices of classifiers (LDA, QDA, SVM, DWD).
% Supported options:
% 1. Binary and Multicategory: LDA-MV, LDA-WV, QDA-MV, QDA-WV
% 2. Binary only: SVM-MV, DWD-MV
%
% delta = setclassmulti(X, sIND, y) returns the LDA rule trained
% by the dataset (X,sIND, y). 'delta' is a cell array with K (number of
% classes) elements.
%
% [delta, pred, errorrate] = setclassmulti(X, sIND, y,
% paramstruct) returns predicted label for testing data and errorrate, see
% below for paramstruct.
%
%
% Input: paramstruct - a Matlab structure of input parameters
%                    Use: "help struct" and "help datatypes" to
%                         learn about these.
%                    Create one, using commands of the form:
%
%        paramstruct = struct('field1',values1,...
%                             'field2',values2,...
%                             'field3',values3) ;
%
%                          where any of the following can be used,
%                          these are optional, misspecified values
%                          revert to defaults
%
%    fields           values
%
%    classifier       'LDA'(default)
%                     'QDA'
%                     'SVM'
%                     'DWD'
%                     'MDEB'
%                     'YA'
%
%    strategy         'WV' (default) use the weighted voting
%                     'MV'   use the majority voting
%
%    classifieroption numeric (default 0.01) any additional parameter for
%                     classifier.
%
%    TestData         Test Data set (similar structure to X)
%
%    Testsetindex     Test Set index (similar structure to sIND)
%
%    TestLabel        Test group labels (similar structure to y)
%
%
%
% By Sungkyu Jung, June-Nov 2013.
%
% See also SetFeat, SetFeatMap, setFeatClass, SetFeatPredict

if nargin < 3
    help setclassmulti.m; return;
end

% set default parameters
classifier = 'LDA';
strategy = 'WV';
classifieroption = 0.01; % L2 penalty for LDA
ireturn_errorrate = 0;
ireturn_pred = 0;
pred = [];
errorrate = [];

if nargin > 3
    
    if isfield(paramstruct,'classifier') ;    %  then change to input value
        classifier = paramstruct.classifier;
    end ;
    if isfield(paramstruct,'strategy') ;    %  then change to input value
        strategy = paramstruct.strategy;
    end ;
    
    if isfield(paramstruct,'classifieroption') ;    %  then change to input value
        classifieroption = paramstruct.classifieroption;
    end ;
    
    if isfield(paramstruct,'TestData') && isfield(paramstruct,'Testsetindex');    %  then change to input value
        datacolN = paramstruct.TestData;
        sINDN = paramstruct.Testsetindex;
        ireturn_pred = 1;
    end ;
    
    if isfield(paramstruct,'TestLabel');    %  then change to input value
        ynew = paramstruct.TestLabel;
        ireturn_errorrate = 1;
    end ;
end
uniquesetindex=unique(sIND);
N = length(uniquesetindex);
[d, Nn] = size(datacol);
yind = zeros(1,Nn);
for i = 1:N
    yind(sIND == uniquesetindex(i)) = y(i);
end

label = unique(y);
K = length(label);

switch classifier
    case 'LDA'
        X = datacol';
        nvec = zeros(K,1); % K-vector of class sizes
        muvec = zeros(d,K); % each column is the class-wise mean
        Sw = zeros(d,d); %With-in (common) covariance matrix
        for k = 1:K;
            nvec(k) = sum(yind==label(k));
            muvec(:,k) = mean(X(yind==label(k),:));
            Sw = Sw + cov(X(yind==label(k),:))*(nvec(k)-1)/Nn;
        end
        Sw = Sw+classifieroption*eye(d);
        
        offset = zeros(1,K);
        ldadir = zeros(d,K);
        for k = 1:K;
            ldadir(:,k) =  pinv(Sw)*muvec(:,k);
            offset(k) = muvec(:,k)'*ldadir(:,k);
        end
        
        delta = cell(K,1);
        for k = 1:K
            delta{k} = @(x) x'*ldadir(:,k) - (.5 * offset(k)) +log(nvec(k));
        end
        
    case 'QDA'
        X = datacol';
        nvec = zeros(K,1); % K-vector of class sizes
        muvec = zeros(d,K); % each column is the class-wise mean
        Sigmaarry = zeros(d,d,K); % covariance matrix for each class
        invSigmaarry = zeros(d,d,K); % covariance matrix for each class
        sumlogevalue = zeros(K,1);
        for k = 1:K
            nvec(k) = sum(yind==label(k));
            muvec(:,k) = mean(X(yind==label(k),:));
            Sigmaarry(:,:,k) = cov(X(yind==label(k),:))+classifieroption*eye(d);
            sumlogevalue(k)=sum(log(eig(Sigmaarry(:,:,k))));
            if isinf(sumlogevalue(k))
                disp('Warning: Nonsingular covariance matrix: increase the sample size or add L2penalty');
                disp('       : Adding 1e-5 for numerical stability')
                Sigmaarry(:,:,k) = Sigmaarry(:,:,k) + (1e-5) * eye(d);
                sumlogevalue(k)=sum(log(eig(Sigmaarry(:,:,k))));
            end
            invSigmaarry(:,:,k) = inv(Sigmaarry(:,:,k));
        end
        
        delta = cell(K,1);
        for k = 1:K
            delta{k} = @(x) -1/2*sumlogevalue(k) ...
                - 1/2*diag((x-repmat(muvec(:,k),1,size(x,2)))' ...
                * (invSigmaarry(:,:,k) * ...
                (x-repmat(muvec(:,k),1,size(x,2))) ))...
                + log(nvec(k));
        end
         
        
    case 'SVM'
        X = datacol;
        [w, b] = SVM1XQ(X(:,yind==label(1)),X(:,yind==label(2)));
        delta = cell(2,1);
        delta{1} = @(x) (w'*x + b > 0)' ;
        delta{2} = @(x) (w'*x + b <= 0)' ;
        
    case 'DWD' 
        X = datacol;
        [w, b] = DWD2XQ(X(:,yind==label(1)),X(:,yind==label(2)));
        delta = cell(2,1);
        delta{1} = @(x) (w'*x + b > 0)' ;
        delta{2} = @(x) (w'*x + b <= 0)' ;
        
    case 'MDEB'
        X = datacol';
        nvec = zeros(K,1); % K-vector of class sizes
        muvec = zeros(d,K); % each column is the class-wise mean
        Sw = zeros(d,d); %With-in (common) covariance matrix
        for k = 1:K;
            nvec(k) = sum(yind==label(k));
            muvec(:,k) = mean(X(yind==label(k),:));
            Sw = Sw + cov(X(yind==label(k),:))*(nvec(k)-1)/Nn;
        end
        
        Sw = Sw+(trace(Sw)/min(Nn,d))*eye(d);
        
        offset = zeros(1,K);
        ldadir = zeros(d,K);
        for k = 1:K;
            ldadir(:,k) =  pinv(Sw)*muvec(:,k);
            offset(k) = muvec(:,k)'*ldadir(:,k);
        end
        
        
        delta = cell(K,1);
        for k = 1:K
            delta{k} = @(x) x'*ldadir(:,k) - (.5 * offset(k)) +log(nvec(k));
        end
        
    case 'YA'
        X = datacol';
        nvec = zeros(K,1); % K-vector of class sizes
        muvec = zeros(d,K); % each column is the class-wise mean
        Sw = zeros(d,d); %With-in (common) covariance matrix
        for k = 1:K;
            nvec(k) = sum(yind==label(k));
            muvec(:,k) = mean(X(yind==label(k),:));
            Sw = Sw + cov(X(yind==label(k),:))*(nvec(k)-1)/Nn;
        end
        [V,D] = eig(Sw);
        lambda = sort(diag(D),'descend');
        omega = sum(lambda)/max([sqrt(d)*Nn^(1/4), min(Nn,d)]);
        lambdanew = omega*zeros(d,1);
        for j = 1:min(d,(N-1));
            lambdanew(j) = max(lambda(j), omega* (lambda(j)/(lambda(j)-sum(lambda(j+1:end))/(N-j))));
        end
        Sw = V*diag(lambdanew)*V';
        
        offset = zeros(1,K);
        ldadir = zeros(d,K);
        for k = 1:K;
            ldadir(:,k) =  pinv(Sw)*muvec(:,k);
            offset(k) = muvec(:,k)'*ldadir(:,k);
        end
        
        delta = cell(K,1);
        for k = 1:K
            delta{k} = @(x) x'*ldadir(:,k) - (.5 * offset(k)) +log(nvec(k));
        end
        
end


if ireturn_pred == 1
    
    [d, Nnew] = size(datacolN);
    
    a = zeros(Nnew,K);
    for k = 1:K
        a(:,k) = delta{k}(datacolN);
    end
    
    uniquesetindex=unique(sINDN);
    Ntest = length(uniquesetindex);
    pred = zeros(Ntest,1);
    
    switch strategy
        case 'WV'
            for j =1:Ntest
                [tmp, maxforsetj] = max(sum(a(sINDN == uniquesetindex(j),:)));
                pred(j) = maxforsetj;
            end
        case 'MV'
            for j =1:Ntest
                [tmp, ind]=max(a(sINDN == uniquesetindex(j),:),[],2);
                votes = zeros(K,1);
                for k = 1:K
                    votes(k) = sum(ind == k);
                end
                [tmp, majorityvote]=max(votes);
                pred(j) = majorityvote;
            end
            
        otherwise
            
    end
    
    
    pred = label(pred);
    if ireturn_errorrate == 1
        
        confusionmat = zeros(K);
        for krow = 1:K
            for kcol = 1:K
                confusionmat(krow,kcol) = sum((ynew == label(krow)) & (pred == label(kcol)));
            end
        end
        errorrate =  1 - (trace(confusionmat)/Ntest);
        
    end
end


